/* *****************************************************************
 * ITS 332 Information Technology II (Networking) Lab
 * Semester 2, 2010
 * SIIT
 *
 * Client/Server Programming Lab
 * File: server.c
 * Date: 24 Jan 2007
 * Version: 1.0
 *
 * Description: 
 * Server to demonstrate TCP sockets programming
 *
 * Usage:
 * server server_port_number
 *
 * Acknowledgement:
 * This code is based on the examples and descriptions from the 
 * Computer Science Department, Rensselaer Polytechnic Institute at:
 * http://www.cs.rpi.edu/courses/sysprog/sockets/sock.html
 *
 * ***************************************************************** */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>

/* ================================================================= */
/* Function Prototypes                                               */
/* ================================================================= */ 
void dostuff(int);

/* ================================================================= */
/* error: display an error message and exit                          */
/* ================================================================= */ 
void error(char *msg)
{
    perror(msg);
    exit(1);
}

/* ================================================================= */
/* main: listen for connections, and create new process for each     */
/* connection. Receive the message from client and acknowledge.      */
/* ================================================================= */ 
int main(int argc, char *argv[])
{
     /* file descriptors that contain values return from socket and 
    and accept system calls */
     int sockfd, newsockfd;
     int portno; /* port number on which server accepts connections */
     int pid; /* process ID for newly created child process */
     /* sockaddr_in is a structure containing an IP address - it is 
        defined in netinet/in.h */
     struct sockaddr_in serv_addr, cli_addr;
     size_t clilen; /* size of the address of the client */

     /* The user must pass the port number that the server listens on
        as a command line parameter (otherwise error) */
     if (argc < 2) {
         fprintf(stderr,"ERROR, no port provided\n");
         exit(1);
     }

     /* First we must create a socket using the socket() system call.
        The three parameters are:
         - address domain of the socket. It may be an Unix socket, an 
           Internet socket or others. We use the Internet socket which
           is defined by the constant AF_INET
         - socket type. Stream (TCP), or Datagram (UDP, SOCK_DGRAM) or
           a raw socket (for accessing IP directly).
         - the protocol. 0 means the operating system will choose the 
           most appropriate protocol: TCP for stream and UDP for 
           datagram. 
        The call to socket returns a file descriptor (or -1 if it fails) */
     sockfd = socket(AF_INET, SOCK_STREAM, 0);
     if (sockfd < 0) 
        error("ERROR opening socket");

     /* bzero sets all values in a buffer to 0. Here we set the server 
        address to 0 */
     bzero((char *) &serv_addr, sizeof(serv_addr));

     /* Get the port number that the user entered via command line */
     portno = atoi(argv[1]);

     /* Now we set the server address in the structure serv_addr */
     /* Note that INADDR_ANY is a constant that refers to the IP
        address of the machine the server is running on. */
     /* Note that the port number must be specified in network byte order.
        Different computer systems represents bytes in different order:
         big endian - most significant bit of byte is first
         little endian - least significant bit of byte is first
        For this reason, everything must be converted to network byte
        order (which is big endian). htons does this conversion. */
     serv_addr.sin_family = AF_INET; /* Protocol family: Internet */
     serv_addr.sin_addr.s_addr = INADDR_ANY; /* Server address */
     serv_addr.sin_port = htons(portno); /* Port number */

     /* The bind() system call binds a socket to an address. This
        may fail if for example the port number is already being 
        used on this machine. */
     if (bind(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0) 
              error("ERROR on binding");

     /* The listen() system call tells the process to listen on the
        socket for connections. The first parameter is a file descriptor
        for the socket and the second parameter is the number of 
        connections that can be queued while the process is handling this
        connection. 5 is a reasonable value for most systems */
     listen(sockfd,5);
     clilen = sizeof(cli_addr);

     /* Now we enter an infinite loop, waiting for connections from clients.
        When a connection is established, we will create a new child process
        using fork(). The child process will handle the data transfer with
        the client. The parent process will wait for another connection. */
     while (1) {
     /* The accept() system call causes the process to block until a
            client connects with the server. The process will wake up once
            the connection has been established (e.g. TCP handshake). 
            The parameters to accept are:
             - the file descriptor of the socket we are waiting on
             - a structure to store the address of the client that connects
             - a variable to store the length of the client address
            It returns a new file descriptor for the socket, and all
            communication is now done with this new descriptor. */
         newsockfd = accept(sockfd, 
               (struct sockaddr *) &cli_addr, &clilen);
         if (newsockfd < 0) 
             error("ERROR on accept");

     /* Create child process to handle the data transfer */
         pid = fork();
         if (pid < 0)
             error("ERROR on fork");
         /* The process ID in the child process will be 0. Hence the child
            process will close the old socket file descriptor and then call
            dostuff() to perform the interactions with the client. When 
            complete, the child process will exit. */
         if (pid == 0)  {
             close(sockfd);
             dostuff(newsockfd);
             exit(0);
         }
     /* This is called by the parent process only. It closes the 
            new socket file descriptor, which is not needed by the parent */
         else close(newsockfd);
     } /* end of while */
     return 0; /* we never get here because we are in infinite loop */
}

/* ================================================================= */
/* dostuff: exchange some messages between client and server. There  */
/* is a separate instance of this function for each connection.      */
/* ================================================================= */ 
void dostuff (int sock)
{
   int n;
   char buffer[256];
      
   /* Set a 256 byte buffer to all zeros */
   bzero(buffer,256);

   /* Read from the socket. This will block until there is something for
      it to read in the socket (i.e. after the client has executed a
      write()). It will read either the total number of characters in the
      socket or 255, whichever is less, and return the number of characters
      read. */
   n = read(sock,buffer,255);
   if (n < 0) error("ERROR reading from socket");

   /* Display the message that was received */
   printf("Here is the message: %s\n",buffer);

   /* Write a message to the socket. The third parameter is the size of 
      the message */
   n = write(sock,"I got your message",18);
   if (n < 0) error("ERROR writing to socket");
}
